from typing import List
from pixaris.generation.base import ImageGenerator
from PIL import Image
import os
import requests
import base64
from io import BytesIO
import time


class FluxFillGenerator(ImageGenerator):
    """
    FluxFillGenerator is responsible for generating images using the Flux API,
    specifically the fill model, which needs an image and a mask as input.
    """

    def validate_inputs_and_parameters(
        self,
        dataset: List[dict[str, List[dict[str, Image.Image]]]] = [],
        parameters: list[dict[str, str, any]] = [],
    ):
        """
        Validates the provided dataset and parameters for image generation.

        :param dataset: A list of datasets containing image and mask information.
        :type dataset: List[dict[str, List[dict[str, Image.Image]]]
        :param parameters: A list of dictionaries containing generation parameters.
        :type parameters: list[dict[str, str, any]]
        :raises ValueError: If the validation fails for any reason (e.g., missing fields).
        """
        # Validate dataset
        if not dataset:
            raise ValueError("Dataset cannot be empty.")

        for entry in dataset:
            if not isinstance(entry, dict):
                raise ValueError("Each entry in the dataset must be a dictionary.")

        # Validate parameters, if given
        if parameters:
            for param in parameters:
                if not isinstance(param, dict):
                    raise ValueError("Each parameter must be a dictionary.")

    def _encode_image_to_base64(self, pillow_image: Image.Image) -> str:
        """
        Encodes a PIL image to a base64 string.

        :param pillow_image: The PIL image.
        :type pillow_image: PIL.Image.Image

        :return: Base64 encoded string representation of the image.
        :rtype: str
        """
        buffered = BytesIO()
        # assigning Image format or JPEG as default
        format = pillow_image.format or "JPEG"
        pillow_image.save(buffered, format=format)
        image_data = buffered.getvalue()
        base64_encoded_string = base64.b64encode(image_data).decode("utf-8")
        return base64_encoded_string

    def _run_flux(
        self, pillow_images: List[dict], generation_params: List[dict]
    ) -> Image.Image:
        """
        Generates images using the Flux API and checks the status until the image is ready.

        :param pillow_images: A list of dictionaries containing pillow images and mask images.
          Example::

          [{'node_name': 'Load Input Image', 'pillow_image': <PIL.Image>}, {'node_name': 'Load Mask Image', 'pillow_image': <PIL.Image>}]
        :type pillow_images: List[dict]
        :param generation_params: A list of dictionaries containing generation params.
        :type generation_params: list[dict]

        :return: The generated image.
        :rtype: PIL.Image.Image
        """
        input_image = pillow_images[1]["pillow_image"]
        mask_image = pillow_images[0]["pillow_image"]

        api_key = os.environ.get("BFL_API_KEY")

        # Set up basis payload
        payload = {
            "image": self._encode_image_to_base64(input_image),
            "mask": self._encode_image_to_base64(mask_image),
            "prompt": "A beautiful landscape with a sunset",
            "steps": 50,
            "prompt_upsampling": False,
            "seed": 1,
            "guidance": 60,
            "output_format": "jpeg",
            "safety_tolerance": 2,
        }

        # Replace generation parameters in the payload
        for param in generation_params:
            payload[param["node_name"]] = param["value"]

        headers = {"Content-Type": "application/json", "X-Key": api_key}

        # Generate image
        response = requests.post(
            "https://api.us1.bfl.ai/v1/flux-pro-1.0-fill", json=payload, headers=headers
        )
        response.raise_for_status()
        request_id = response.json()["id"]

        # Check image status
        status_url = "https://api.us1.bfl.ai/v1/get_result"

        while True:
            time.sleep(1)
            status_response = requests.get(
                status_url,
                headers={"accept": "application/json", "x-key": api_key},
                params={"id": request_id},
            )
            status_response.raise_for_status()
            result = status_response.json()

            if result["status"] == "Ready":
                image_url = result["result"]["sample"]
                break
            print(f"Status: {result['status']}")

        # Download and return the image
        image_response = requests.get(image_url)
        image_response.raise_for_status()

        return Image.open(BytesIO(image_response.content))

    def generate_single_image(self, args: dict[str, any]) -> tuple[Image.Image, str]:
        """
        Generates a single image based on the provided arguments.

        :param args: A dictionary containing the following keys:
        * pillow_images (list[dict]): A list of dictionaries containing pillow images and mask images.
        * generation_params (list[dict]): A list of dictionaries containing generation params.
        :type args: dict[str, any]

        :return: A tuple containing:
        * image (Image.Image): The generated image.
        * image_name (str): The name of the generated image.
        :rtype: tuple[Image.Image, str]
        """
        pillow_images = args.get("pillow_images", [])
        generation_params = args.get("generation_params", [])

        image = self._run_flux(pillow_images, generation_params)

        # Since the names should all be the same, we can just take the first.
        image_name = pillow_images[0]["pillow_image"].filename.split("/")[-1]

        return image, image_name



from functools import partial
import time
from decorator import decorator

# Inspired by https://github.com/invl/retry/blob/master/retry/api.py, but without the py dependency


def __retry_internal(
    f, exceptions=Exception, tries=-1, delay=0, max_delay=None, backoff=1
):
    """
    Executes a function and retries it if it failed.

    :param f: the function to execute.
    :type f: callable
    :param exceptions: an exception or a tuple of exceptions to catch. default: Exception.
    :type exceptions: BaseException
    :param tries: the maximum number of attempts. default: -1 (infinite).
    :type tries: int
    :param delay: initial delay between attempts. default: 0.
    :type delay: int
    :param max_delay: the maximum value of delay. default: None (no limit).
    :type max_delay: int
    :param backoff: multiplier applied to delay between attempts. default: 1 (no backoff).
    :type backoff: int
    :returns: the result of the f function.
    :rtype: any
    """
    _tries, _delay = tries, delay
    while _tries:
        try:
            return f()
        except exceptions:
            _tries -= 1
            if not _tries:
                raise

            time.sleep(_delay)
            _delay *= backoff

            if max_delay is not None:
                _delay = min(_delay, max_delay)


def retry(exceptions=Exception, tries=-1, delay=0, max_delay=None, backoff=1):
    """Returns a retry decorator.

    :param exceptions: an exception or a tuple of exceptions to catch. default: Exception.
    :type exceptions: BaseException
    :param tries: the maximum number of attempts. default: -1 (infinite).
    :type tries: int
    :param delay: initial delay between attempts. default: 0.
    :type delay: int
    :param max_delay: the maximum value of delay. default: None (no limit).
    :type max_delay: int
    :param backoff: multiplier applied to delay between attempts. default: 1 (no backoff).
    :type backoff: int
    :return: a retry decorator.
    :rtype: decorator
    """

    @decorator
    def retry_decorator(f, *fargs, **fkwargs):
        args = fargs if fargs else list()
        kwargs = fkwargs if fkwargs else dict()
        return __retry_internal(
            partial(f, *args, **kwargs), exceptions, tries, delay, max_delay, backoff
        )

    return retry_decorator



from typing import List
from pixaris.generation.base import ImageGenerator
from pixaris.generation.comfyui_utils.workflow import ComfyWorkflow
from PIL import Image
import hashlib


class ComfyGenerator(ImageGenerator):
    """
    ComfyGenerator is a class that extends the ImageGenerator class to provide functionality for generating images using a specified workflow and API host.
    It uses the ComfyUI API to generate images based on the provided workflow and input parameters.

    :param workflow_apiformat_json: The workflow file in API format.
    :type workflow_apiformat_json: dict
    :param api_host: The API host URL. For local experimenting, put "localhost:8188".
      There has to be a tunnel to a running comfyUI instance active on port 8188
    :type api_host: str
    """

    def __init__(
        self,
        workflow_apiformat_json: dict,
        api_host: str = "localhost:8188",
    ):
        self.api_host = api_host
        self.workflow_apiformat_json = workflow_apiformat_json
        self.workflow = ComfyWorkflow(
            api_host=self.api_host,
            workflow_apiformat_json=self.workflow_apiformat_json,
        )

    def _get_unique_int_for_image(self, pillow_image: Image.Image) -> int:
        """
        Gets a unique int for an image calculated from image name and hash. This is needed to have a unique
        seed for the experiments but have the same seed for the same image in different experiments.

        :param pillow_image: The PIL image.
        :type pillow_image: Image.Image
        :return: The unique integer for the image.
        :rtype: int
        """
        file_name = pillow_image.filename.split("/")[-1].split(".")[0]
        img_bytes = file_name.encode("utf-8") + pillow_image.tobytes()
        img_hash = hashlib.md5(img_bytes).hexdigest()
        unique_number = int(img_hash, 16)
        final_seed = (unique_number % 1000000) + 1  # cannot be too big for comfy
        return final_seed

    def validate_inputs_and_parameters(
        self,
        dataset: List[dict[str, List[dict[str, Image.Image]]]] = [],
        parameters: list[dict[str, str, any]] = [],
    ) -> str:
        """
        Validates the workflow file to ensure that it is in the correct format.

        :param dataset: A list of dictionaries containing the images to be loaded.
        :type dataset: List[dict[str, List[dict[str, Image.Image]]]
        :param parameters: A list of dictionaries containing the parameters to be used for the image generation process.
        :type parameters: list[dict[str, str, any]]
        :return: The path to the validated workflow file.
        :rtype: str
        """

        # assert each existing element of generation_params has the keys "node_name", "input", "value"
        for value_info in parameters:
            if not all(key in value_info for key in ["node_name", "input", "value"]):
                raise ValueError(
                    "Each generation_param dictionary should contain the keys 'node_name', 'input', and 'value'."
                )

        # assert the params can be applied to the workflow
        self.workflow.check_if_parameters_are_valid(parameters)

        # assert each element of pillow_images has the keys "pillow_image", "node_name"
        for image_info in dataset:
            image_set = image_info.get("pillow_images", [])
            # check if "node_name", "pillow_image" are keys in image_info
            if not all(
                key in image_dict
                for image_dict in image_set
                for key in ["node_name", "pillow_image"]
            ):
                raise ValueError(
                    "Each pillow_images dictionary should contain the keys 'node_name' and 'pillow_image'."
                )
            # check if all pillow_images are PIL Image objects
            if not all(
                isinstance(image_dict["pillow_image"], Image.Image)
                for image_dict in image_set
            ):
                wrong_types = [
                    type(image_dict["pillow_image"])
                    for image_dict in image_set
                    if not isinstance(image_dict["pillow_image"], Image.Image)
                ]
                raise ValueError(
                    "All pillow_images should be PIL Image objects. Got: ", wrong_types
                )

    def _modify_workflow(
        self,
        pillow_images: list[dict[str, Image.Image]] = [],
        generation_params: list[dict[str, str, any]] = [],
    ) -> ComfyWorkflow:
        """
        Modifies the workflow to generate a single image based on the provided arguments.

        :param pillow_images: A list of dictionaries containing the images to be loaded.
        :type pillow_images: list[dict[str, Image.Image]], optional
        :param generation_params: A list of dictionaries containing the parameters to be used for the image generation process.
        :type generation_params: list[dict[str, str, any]], optional
        :return: The modified workflow.
        :rtype: ComfyWorkflow
        """
        self.workflow.adjust_workflow_to_generate_one_image_only()

        # adjust all generation_params
        if generation_params:
            self.workflow.set_generation_params(generation_params)

        # Load and set images from pillow_images
        for image_info in pillow_images:
            input_image = image_info["pillow_image"]
            self.workflow.set_image(image_info["node_name"], input_image)

        # set seed or warn if it is not being set.
        if self.workflow.check_if_node_exists(
            "KSampler (Efficient) - Generation"
        ):  # not present e.g. in mask workflows
            self.workflow.set_value(
                "KSampler (Efficient) - Generation",
                "seed",
                self._get_unique_int_for_image(pillow_images[0]["pillow_image"]),
            )
        else:
            print(
                "Node 'KSampler (Efficient) - Generation' not found in the workflow. Seed will not be set."
            )

        return self.workflow

    def generate_single_image(self, args: dict[str, any]) -> tuple[Image.Image, str]:
        # Todo: change the docstring format when this issue is closed: https://github.com/sphinx-doc/sphinx/issues/4220
        """
        Generates a single image based on the provided arguments. For this it modifies and executed the workflow to generate the image.

        :param args: A dictionary containing the following keys:
        * "workflow_apiformat_json" (dict): The workflow file in JSON apiformat.
        * "pillow_images" (list[dict]): A dict of [str, Image.Image].
          The keys should be Node names
          The values should be the PIL Image objects to be loaded.
          Should look like this::

          "pillow_images": [{
          "node_name": "Load Input Image",
          "pillow_image": Image.new("RGB", (100, 100), color="red")}]
        * "generation_params" (list[dict]): A dictionary of generation_params for the image generation process.
          Should look like this::

          "generation_params": [{
          "node_name": "GroundingDinoSAMSegment (segment anything)",
          "input": "prompt",
          "value": "model, bag, hair"}]

        :rtype args: dict[str, any]
        :return: The generated image and its name
        :rtype: tuple[Image.Image, str]
        """

        assert "workflow_apiformat_json" in args, (
            "The key 'workflow_apiformat_json' is missing."
        )

        pillow_images = args.get("pillow_images", [])
        generation_params = args.get("generation_params", [])

        # since the names should all be the same, we can just take the first.
        image_name = pillow_images[0]["pillow_image"].filename.split("/")[-1]

        self.workflow = self._modify_workflow(
            pillow_images=pillow_images,
            generation_params=generation_params,
        )

        try:
            self.workflow.execute()
            image = self.workflow.get_image("Save Image")[0]
            return image, image_name
        except ConnectionError as e:
            print(
                "Connection Error. Did you forget to build the iap tunnel to ComfyUI on port 8188?"
            )
            raise e



from io import BytesIO
from PIL import Image


def extract_value_from_list_of_dicts(
    dict_list,
    identifying_key,
    identifying_value,
    return_key,
    default_value=None,
):
    """
    Extracts a value from a list of dictionaries based on a key-value pair.
    This function searches through a list of dictionaries and returns the value associated with a specified key
    for the first dictionary that matches a given key-value pair. If no such dictionary is found, it returns a default value.
    If the default value is not provided and no matching dictionary is found, a ValueError is raised.

    :param dict_list: A list of dictionaries to search through.
    :type dict_list: list[dict]
    :param itentifying_key: The key to identify the dictionary.
    :type itentifying_key: _any
    :param identifying_value: The value to match against the identifying key.
    :type identifying_value: _any
    :param return_key: The key whose value is to be returned from the matching dictionary.
    :type return_key: _any
    :param default_value: The value to return if no matching dictionary is found. If not provided, defaults to None.
    :type default_value: _any, optional
    :raises ValueError: If no matching dictionary is found and no default value is provided.
    :return: The value associated with the return key from the matching dictionary, or the default value if no match is found.
    :rtype: _any
    """
    return_value = next(
        (
            param[return_key]
            for param in dict_list
            if param[identifying_key] == identifying_value
        ),
        default_value,
    )
    if return_value is None:
        raise ValueError(
            f"No dict with pair '{identifying_key}': '{identifying_value}' and key '{return_key}' found."
        )
    return return_value


def encode_image_to_bytes(pillow_image: Image.Image) -> bytes:
    """
    Encodes a PIL image to bytes.

    :param pillow_image: The PIL image.
    :type pillow_image: PIL.Image.Image

    :return: Byte array representation of the image.
    :rtype: bytes
    """
    imgByteArr = BytesIO()
    pillow_image.save(imgByteArr, format=pillow_image.format)
    imgByteArr = imgByteArr.getvalue()
    return imgByteArr



import gradio as gr
from pixaris.experiment_handlers.base import ExperimentHandler
import pandas as pd

PROJECTS = []


def render_experiment_tracking_tab(
    experiment_handler: ExperimentHandler,
    results_directory: str,
):
    dataset_experiment_tracking_results = gr.State(pd.DataFrame())
    with gr.Sidebar(open=True, position="right"):
        gr.Markdown("Experiments")
        with gr.Row(scale=8):
            # load all projects and corresponding datasets at the beginning
            PROJECTS_DICT = experiment_handler.load_projects_and_datasets()
            global PROJECTS
            PROJECTS = [""] + list(PROJECTS_DICT.keys())

            project_name = gr.Dropdown(
                choices=PROJECTS,
                label="Project",
                filterable=True,
            )

            # initialise hidden dataset and define updating function
            dataset = gr.Dropdown(visible=False)

            def update_dataset_choices(project_name, dataset):
                """Update choices of feedback iterations for selected project and display reload button."""
                dataset_choices = PROJECTS_DICT[project_name]

                dataset = gr.Dropdown(
                    label="Dataset",
                    choices=dataset_choices,
                    filterable=True,
                    multiselect=False,
                    interactive=True,
                    visible=True,
                )
                return dataset

            # change dataset choices upon project change
            project_name.change(
                fn=update_dataset_choices,
                inputs=[
                    project_name,
                    dataset,
                ],
                outputs=[dataset],
            )

            # initialise hidden experiments and define updating function
            experiments = gr.Dropdown(visible=False)

            def update_experiments_choices_and_load_table(
                project_name, dataset, experiments, dataset_experiment_tracking_results
            ):
                """Update choices of feedback iterations for selected project and display reload button."""
                dataset_experiment_tracking_results = (
                    experiment_handler.load_experiment_results_for_dataset(
                        project=project_name,
                        dataset=dataset,
                    )
                )
                experiment_choices = list(
                    dataset_experiment_tracking_results["experiment_run_name"]
                    .dropna()
                    .unique()
                )
                experiment_choices.sort()
                experiments = gr.Dropdown(
                    choices=experiment_choices,
                    label="Experiments",
                    filterable=True,
                    multiselect=True,
                    max_choices=100,
                    visible=True,
                )
                return experiments, dataset_experiment_tracking_results

            # change experiments choices and load table upon dataset change
            dataset.change(
                fn=update_experiments_choices_and_load_table,
                inputs=[
                    project_name,
                    dataset,
                    experiments,
                    dataset_experiment_tracking_results,
                ],
                outputs=[experiments, dataset_experiment_tracking_results],
            )
        with gr.Row(scale=1):
            reload_projects_button = gr.Button(
                "Reload projects",
                variant="secondary",
                interactive=True,
                size="sm",
            )

            def reload_projects(project_name):
                PROJECTS_DICT = experiment_handler.load_projects_and_datasets()
                global PROJECTS
                PROJECTS = [""] + list(PROJECTS_DICT.keys())
                return project_name

            reload_projects_button.click(
                fn=reload_projects,
                inputs=[project_name],
                outputs=[project_name],
            )
        # columns and gallery height sliders
        with gr.Row(scale=1):
            columns = gr.Slider(
                minimum=1,
                maximum=20,
                value=8,
                label="Number of images per row",
                step=1,
            )
        with gr.Row(scale=1):
            gallery_height = gr.Slider(
                minimum=100,
                maximum=1000,
                value=200,
                label="Gallery height",
                step=10,
            )

    with gr.Tab("Images"):
        # Display images in a gallery

        @gr.render(inputs=[project_name, dataset, experiments, columns, gallery_height])
        def render_images_per_experiment(
            project_name, dataset, experiments, columns, gallery_height
        ):
            """
            Renders one accordion with a gallery per experiment. Render decorator enables listening to experiments checkbox group.
            """
            if not experiments:
                gr.Markdown("No experiment selected.")
                return
            for experiment_name in experiments:
                with gr.Accordion(label=f"Experiment {experiment_name}"):
                    experiment_images = experiment_handler.load_images_for_experiment(
                        project=project_name,
                        dataset=dataset,
                        experiment_run_name=experiment_name,
                        local_results_directory=results_directory,
                    )
                    gr.Gallery(
                        value=experiment_images,
                        columns=columns,
                        rows=1,
                        show_download_button=True,
                        show_fullscreen_button=True,
                        height=gallery_height,
                        object_fit="contain",
                    )

    with gr.Tab("Table"):
        # Display experiment results in a table

        @gr.render(inputs=[experiments, dataset_experiment_tracking_results])
        def render_experiment_results_table(
            experiments, dataset_experiment_tracking_results
        ):
            if not experiments:
                gr.Markdown("No experiment selected.")
                return

            table_data = dataset_experiment_tracking_results.copy()
            table_data = table_data.loc[
                dataset_experiment_tracking_results["experiment_run_name"].isin(
                    experiments
                )
            ]
            gr.DataFrame(
                table_data,
                label="Experiment Results",
                wrap=True,
                show_search="filter",
                max_height=1000,
            )



from abc import abstractmethod
import pandas as pd
from datetime import datetime


class FeedbackHandler:
    feedback_iteration_choices = []
    feedback_df = pd.DataFrame()

    @abstractmethod
    def write_single_feedback(
        self,
        feedback: dict[str, any],
    ) -> None:
        """
        Writes feedback for one image to the feedback table.

        :param feedback: dict with feedback information. Dict is expected to have the following keys:
        * project: name of the project
        * feedback_iteration: name of the iteration
        * dataset: name of the evaluation set (optional)
        * image_name: name of the image
        * experiment_name: name of the experiment (optional)
        * feedback_indicator: string with feedback value (either "Like", "Dislike", or "Neither")
        * comment: string with feedback comment (optional)
        :type feedback: dict
        """
        pass

    def _construct_feedback_row_to_insert(self, feedback: dict[str, any]) -> None:
        """
        Constructs a row to insert to the feedback table based on the provided feedback dictionary.

        :param feedback: dict with feedback information. Dict is expected to have the following keys:
        * project: name of the project
        * feedback_iteration: name of the iteration
        * dataset: name of the evaluation set (optional)
        * image_name: name of the image
        * experiment_name: name of the experiment (optional)
        * feedback_indicator: string with feedback value (either "Like", "Dislike", or "Neither")
        * comment: string with feedback comment (optional)
        :type feedback: dict
        """
        # assert non-nullable values are present
        assert all(
            key in feedback.keys()
            for key in [
                "project",
                "feedback_iteration",
                "image_name",
                "feedback_indicator",
            ]
        ), (
            "Missing required feedback keys. Must have 'project', 'feedback_iteration', 'image_name', 'feedback_indicator'"
        )

        # setup row to insert to table
        row_to_insert = {
            "project": feedback["project"],
            "feedback_iteration": feedback["feedback_iteration"],
            "dataset": feedback.get("dataset", ""),
            "image_name": feedback["image_name"],
            "experiment_name": feedback.get("experiment_name", ""),
            "date": datetime.now().isoformat(),
            "comment": feedback.get("comment", ""),
            "misc": feedback.get("misc", ""),
        }

        # determine what to write to feedback columns
        feedback_indicator = feedback["feedback_indicator"]

        row_to_insert["likes"] = 0
        row_to_insert["dislikes"] = 0

        if feedback_indicator == "Like":
            row_to_insert["likes"] = 1
        elif feedback_indicator == "Dislike":
            row_to_insert["dislikes"] = 1
        elif (
            feedback_indicator
            != "Neither"  # Neither is used when uploading images (no feedback given)
        ):
            raise ValueError(
                "Invalid feedback indicator. Must be 'Like', 'Dislike', or 'Neither'"
            )
        return row_to_insert

    @abstractmethod
    def create_feedback_iteration(
        self,
        local_image_directory: str,
        project: str,
        feedback_iteration: str,
        date_suffix: str = None,
        dataset: str = None,
        experiment_name: str = None,
    ):
        """
        Creates a feedback iteration, which involves persisting the creation to a feedback table as well as
        save the corresponding images.

        :param images_directory: Path to local directory containing images to upload
        :type images_directory: str
        :param project: Name of the project
        :type project: str
        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        :param date_suffix: Date suffix for versioning. Will be set automatically to today if not provided.
        :type date_suffix: str
        :param dataset: Name of the evaluation dataset (optional)
        :type dataset: str
        :param experiment_name: Name of the experiment (optional)
        :type experiment_name: str
        """
        pass

    @abstractmethod
    def load_projects_list(self) -> list[str]:
        """
        Returns the list of projects.

        :return: List of projects
        :rtype: list[str]
        """
        pass

    def get_feedback_per_image(self, feedback_iteration, image_name) -> dict:
        """
        Get feedback for a specific image.

        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        :param image_name: Name of the image
        :type image_name: str
        :return: Dictionary with feedback information for the image with the following format:
         {"likes": int, "dislikes": int, "comments_liked": list, "comments_disliked": list}
        :rtype: dict
        """
        # happens because of "batching" of image to display all images the same size
        if image_name == "None":
            return {
                "likes": 0,
                "dislikes": 0,
                "comments_liked": [],
                "comments_disliked": [],
            }
        else:
            feedback_info_df = self.feedback_df.query(
                f'feedback_iteration == "{feedback_iteration}" and image_name == "{image_name}"'
            ).reset_index(inplace=False)
            feedback_info_dict = feedback_info_df.loc[
                0, ["likes", "dislikes", "comments_liked", "comments_disliked"]
            ].to_dict()
            return feedback_info_dict

    def load_all_feedback_iteration_info_for_project(self, project: str) -> None:
        """
        Retrieves feedback data for a project. Adds paths for location of images in
        local directory to the dataframe. Saves the resulting df to self.feedback_df.
        Saves the list of feedback iterations to self.feedback_iteration_choices.

        :param project: Name of the project
        :type project: str
        """
        feedback_df = self._load_feedback_df(project)
        self.feedback_df = feedback_df

        # determine feedback iterations to choose from in this project
        feedback_iteration_choices = feedback_df["feedback_iteration"].unique().tolist()
        feedback_iteration_choices.sort()
        self.feedback_iteration_choices = feedback_iteration_choices

        print(f"Done. Found feedback iterations: {feedback_iteration_choices}")

    @abstractmethod
    def load_images_for_feedback_iteration(
        self,
        feedback_iteration: str,
        sorting: str = "image_name",
    ) -> list[str]:
        """
        Returns list of local image paths that belong to the feedback iteration. Optionally loads images.

        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        :param sorting: Sorting option for the images. Can be "image_name", "likes", or "dislikes". Default is "image_name".
        :type sorting: str
        :return: List of local image paths
        :rtype: list[str]
        """
        pass

    def _sort_iteration_df(
        self,
        iteration_df: pd.DataFrame,
        sorting: str = "image_name",
    ) -> pd.DataFrame:
        """
        Sorts the iteration dataframe based on the specified sorting option.

        :param iteration_df: DataFrame containing feedback data for one iteration.
        :type iteration_df: pd.DataFrame
        :param sorting: Sorting option for the images. Can be "image_name", "likes", or "dislikes". Default is "image_name".
        :type sorting: str
        :return: Sorted DataFrame
        :rtype: pd.DataFrame
        """
        needed_columns = ["image_name", "image_path_local", "likes", "dislikes"]
        assert all(column in iteration_df.columns for column in needed_columns), (
            f"Expected columns not found in DataFrame. Expected {needed_columns}, found {iteration_df.columns}"
        )

        if sorting == "image_name":
            sorted_df = iteration_df.sort_values("image_name")
        elif sorting == "likes":
            sorted_df = iteration_df.groupby("image_path_local")[
                ["likes", "dislikes"]
            ].agg("sum")
            sorted_df = sorted_df.sort_values("likes", ascending=False).reset_index()
        elif sorting == "dislikes":
            sorted_df = iteration_df.groupby("image_path_local")[
                ["likes", "dislikes"]
            ].agg("sum")
            sorted_df = sorted_df.sort_values("dislikes", ascending=False).reset_index()
        else:
            raise ValueError(
                "Invalid sorting option. Must be 'alphabetical', 'likes', or 'dislikes'"
            )
        return sorted_df



import gradio as gr
from pixaris.feedback_handlers.base import FeedbackHandler

PROJECTS = []


def render_feedback_tab(
    feedback_handler: FeedbackHandler,
):
    # initially load all projects
    global PROJECTS
    PROJECTS = [""] + feedback_handler.load_projects_list()

    feedback_details = (
        gr.State(  # is adjusted from inside a gr.render decorated function. See below.
            value={
                "project": "",
                "feedback_iteration": "",
                "image_name": "",
                "feedback_indicator": False,
                "comment": "",
            }
        )
    )

    def adjust_feedback_details(
        img_path: str, feedback_indicator: bool, comment: str, previous_details: dict
    ):
        """
        This function is only here to adjust the feedback details (that are a gr.State object). This is
        necessary because the feedback details are determined within a function that has the gr.render
        decorator. Thus, the only way to trigger a function within the gr.render function (in this case
        we want to write the feedback) is to write the inputs to a gr.State object and then call the
        respective function within here.
        See here https://www.gradio.app/guides/state-in-blocks (see example with cart).
        """
        previous_details["project"] = img_path.split("/")[-4]  # todo make more robust
        previous_details["feedback_iteration"] = img_path.split("/")[-2]
        previous_details["image_name"] = img_path.split("/")[-1]
        previous_details["feedback_indicator"] = feedback_indicator
        previous_details["comment"] = comment
        feedback_handler.write_single_feedback(feedback=previous_details)
        return previous_details

    with gr.Sidebar(open=True, position="right"):
        with gr.Row(scale=8):
            project_name = gr.Dropdown(
                choices=PROJECTS,
                label="Project",
                filterable=True,
            )

            # initialise hidden feedback iterations and define updating function
            feedback_iterations = gr.Dropdown(visible=False)

            def update_feedback_iteration_choices(project_name, feedback_iterations):
                """Update choices of feedback iterations for selected project."""
                feedback_handler.load_all_feedback_iteration_info_for_project(
                    project_name
                )
                feedback_iteration_choices = feedback_handler.feedback_iteration_choices

                feedback_iterations = gr.Dropdown(
                    label="Feedback Iterations",
                    value=None,
                    choices=feedback_iteration_choices,
                    visible=True,
                    filterable=True,
                    multiselect=True,
                    max_choices=100,
                    interactive=True,
                )
                return feedback_iterations

            # update feedback iterations choices upon project change
            project_name.change(
                fn=update_feedback_iteration_choices,
                inputs=[
                    project_name,
                    feedback_iterations,
                ],
                outputs=[feedback_iterations],
            )
        with gr.Row(scale=1):
            reload_projects_button = gr.Button(
                "Reload projects",
                variant="secondary",
                interactive=True,
                size="sm",
            )

            def reload_projects(project_name):
                global PROJECTS
                PROJECTS = feedback_handler.load_projects_list()
                project_name = gr.Dropdown(
                    choices=PROJECTS,
                    label="Project",
                    filterable=True,
                )
                return project_name

            reload_projects_button.click(
                fn=reload_projects,
                inputs=[project_name],
                outputs=[project_name],
            )
        with gr.Row(scale=1):
            columns = gr.Slider(
                minimum=1,
                maximum=7,
                value=5,
                label="Number of images per row",
                step=1,
            )
        with gr.Row(scale=1):
            display_feedback_checkbox = gr.Checkbox(
                value=False,
                label="Show previous feedback",
                interactive=True,
            )
            sorting_of_images = gr.Radio(
                choices=["image_name", "likes", "dislikes"],
                label="Sorting of images",
                value="image_name",
            )
            reload_feedback_button = gr.Button(
                "Reload feedback",
                variant="secondary",
                interactive=True,
                size="sm",
            )
            reloaded_feedback_count = gr.State(value=1)

            def reload_feedback(project_name, reloaded_feedback_count):
                feedback_handler.load_all_feedback_iteration_info_for_project(
                    project_name
                )
                reloaded_feedback_count += 1
                gr.Info("Reloaded feedback.", duration=2)
                return reloaded_feedback_count

            reload_feedback_button.click(
                fn=reload_feedback,
                inputs=[project_name, reloaded_feedback_count],
                outputs=[reloaded_feedback_count],
            )

    @gr.render(
        inputs=[
            feedback_iterations,
            columns,
            display_feedback_checkbox,
            sorting_of_images,
            reloaded_feedback_count,
        ]
    )
    def render_images_per_iteration(
        feedback_iterations,
        columns,
        display_feedback_checkbox,
        sorting_of_images,
        reloaded_feedback_count,
    ):
        """
        This function renders the images for each feedback iteration. It is decorated with gr.render
        to allow for dynamic rendering of the images based on the selected feedback iterations.
        - for each feedback_iteration, there is a separate accordion
        - each accordion contains rows of images with the number of columns specified by the user
        - each image is associated with feedback functionality
        """
        if not feedback_iterations:
            gr.Markdown("No feedback iteration selected.")
            return
        for feedback_iteration in feedback_iterations:
            # load the images corresponding to this feedback iteration
            feedback_iteration_images = (
                feedback_handler.load_images_for_feedback_iteration(
                    feedback_iteration, sorting_of_images
                )
            )

            # split images into batches of number of columns
            num_images = len(feedback_iteration_images)
            images_batches = [
                feedback_iteration_images[i : i + columns]
                for i in range(0, num_images, columns)
            ]

            # fill up last batch with Nones to have same number of images in each row
            if len(images_batches[-1]) < columns:
                images_batches[-1] += [None] * (columns - len(images_batches[-1]))

            # render images
            min_width_elements = "10px"
            with gr.Accordion(label=f"Iteration {feedback_iteration}"):
                for batch in images_batches:
                    with gr.Row(variant="compact"):
                        for img in batch:
                            # string of image_name is needed later on, img will be modified by gradio hereafter.
                            img_path = str(img)
                            img_name = img_path.split("/")[-1]
                            # only display image with buttons if it exists
                            element_visible_bool = bool(img)
                            with gr.Column(
                                variant="compact", min_width=min_width_elements
                            ):
                                gr.Image(
                                    value=img,
                                    label=img_name,
                                    show_download_button=True,
                                    show_fullscreen_button=True,
                                    visible=element_visible_bool,
                                    min_width=min_width_elements,
                                    scale=1,
                                )
                                img_textbox = gr.Textbox(
                                    value=img_path, visible=False
                                )  # needed bc gr.render
                                comment = gr.Textbox(
                                    label="Comment",
                                    value="",
                                    visible=element_visible_bool,
                                    min_width=min_width_elements,
                                    scale=1,
                                    interactive=True,
                                )
                                feedback_indicator = gr.Radio(
                                    choices=["Like", "Dislike"],
                                    label="Feedback",
                                    visible=element_visible_bool,
                                )
                                feedback_button = gr.Button(
                                    "Send feedback for this image",
                                    visible=element_visible_bool,
                                    size="sm",
                                    min_width=min_width_elements,
                                    scale=1,
                                    variant="primary",
                                    interactive=False,
                                )

                                # feedback button is only clickable if feedback indicator is changed recently
                                def change_feedback_button_interactivity(
                                    feedback_indicator,
                                    feedback_button,
                                ):
                                    if feedback_indicator:
                                        feedback_button = gr.Button(interactive=True)
                                    else:
                                        feedback_button = gr.Button(interactive=False)

                                    return feedback_button

                                feedback_indicator.change(
                                    fn=change_feedback_button_interactivity,
                                    inputs=[feedback_indicator, feedback_button],
                                    outputs=[feedback_button],
                                )

                                # adjusts gr.State object with feedback details and writes feedback
                                feedback_button.click(
                                    fn=adjust_feedback_details,
                                    inputs=[
                                        img_textbox,
                                        feedback_indicator,
                                        comment,
                                        feedback_details,
                                    ],
                                    outputs=[feedback_details],
                                ).then(
                                    # change feedback button to non-interactive again (avoid double feedback)
                                    fn=lambda: gr.update(interactive=False),
                                    inputs=None,
                                    outputs=feedback_button,
                                )

                                feedback = feedback_handler.get_feedback_per_image(
                                    feedback_iteration=feedback_iteration,
                                    image_name=img_name,
                                )
                                gr.Markdown(
                                    label="Previous Feedback",
                                    value=f"Likes: {feedback['likes']} - Comments: {feedback['comments_liked']}"
                                    if feedback and reloaded_feedback_count > 0
                                    else "",
                                    visible=display_feedback_checkbox
                                    and element_visible_bool,
                                )
                                gr.Markdown(
                                    label="Previous Feedback",
                                    value=f"Dislikes: {feedback['dislikes']} - Comments: {feedback['comments_disliked']}"
                                    if feedback and reloaded_feedback_count > 0
                                    else "",
                                    visible=display_feedback_checkbox
                                    and element_visible_bool,
                                )



from abc import abstractmethod


class BaseMetric:
    """When implementing a new Metric, inherit from this one and implement all the abstract methods."""

    @abstractmethod
    def calculate(self, x: any) -> dict:
        pass



from typing import Iterable
from PIL import Image
from PIL.PngImagePlugin import PngInfo
from pixaris.experiment_handlers.base import ExperimentHandler
import json
import pandas as pd
from google.cloud import bigquery, storage
import os
import time
from datetime import datetime
import gradio as gr
from pixaris.utils.bigquery import ensure_table_exists


class GCPExperimentHandler(ExperimentHandler):
    """
    GCPExperimentHandler is a class for storing experiment results in Google Cloud Storage and BigQuery.
    It is also used to retrieve the images for display.

    :param gcp_project_id: The Google Cloud Platform project ID.
    :type gcp_project_id: str
    :param gcp_bq_experiment_dataset: The BigQuery dataset for storing experiment results.
    :type gcp_bq_experiment_dataset: str
    :param gcp_pixaris_bucket_name: The name of the Google Cloud Storage bucket for storing images.
    :type gcp_pixaris_bucket_name: str
    :type ExperimentHandler: _type_
    """

    def __init__(
        self,
        gcp_project_id: str,
        gcp_bq_experiment_dataset: str,
        gcp_pixaris_bucket_name: str,
    ):
        self.gcp_project_id = gcp_project_id
        self.gcp_bq_experiment_dataset = gcp_bq_experiment_dataset
        self.gcp_pixaris_bucket_name = gcp_pixaris_bucket_name

        self.storage_client = None
        self.bigquery_client = None
        self.pixaris_bucket = None

        self.project = None
        self.dataset = None
        self.experiment_run_name = None

    def _ensure_unique_experiment_run_name(self) -> str:
        """
        Ensures that the experiment run name is unique by appending a timestamp and random number if necessary.

        :return: A unique experiment run name.
        :rtype: str
        """
        timestamp = datetime.now().strftime("%y%m%d")
        self.experiment_run_name = f"{timestamp}-{self.experiment_run_name}"

        blobs = self.pixaris_bucket.list_blobs(
            prefix=f"results/{self.project}/{self.dataset}"
        )
        for blob in blobs:
            if self.experiment_run_name in blob.name:
                return self.experiment_run_name + datetime.now().strftime("%H%M")
        return self.experiment_run_name

    def _validate_args(self, args: dict[str, any]):
        """
        Validates the arguments passed to the experiment handler.

        :param args: A dictionary of arguments to validate.
        :type args: dict[str, any]
        :raises AssertionError: If the arguments do not meet the required structure.
        """
        # check if all keys are strings
        assert all(isinstance(key, str) for key in args.keys()), (
            "All keys must be strings."
        )

        # check if "pillow_images" is a list of dictionaries containing the correct keys
        if "pillow_images" in args:
            pillow_images = args["pillow_images"]
            assert isinstance(pillow_images, list), "pillow_images must be a list."
            assert all(isinstance(item, dict) for item in pillow_images), (
                "Each item in the list must be a dictionary."
            )
            assert all(
                all(key in item for key in ["node_name", "pillow_image"])
                for item in pillow_images
            ), "Each dictionary must contain the keys 'node_name' and 'pillow_image'."

    def _upload_to_gcs(self, key: str, value: any) -> str:
        """
        Uploads a file (image or JSON) to Google Cloud Storage and returns its GCS path.

        :param key: The key or name of the file.
        :type key: str
        :param value: The content to upload (PIL Image or dict).
        :type value: any
        :return: The GCS path of the uploaded file.
        :rtype: str
        """

        if isinstance(value, Image.Image):
            tmp_path = f"{key}.png"
            metadata = PngInfo()
            for metadata_key, metadata_value in value.info.items():
                metadata.add_text(metadata_key, str(metadata_value))
            value.save(tmp_path, pnginfo=metadata)
        elif isinstance(value, dict):
            tmp_path = f"{key}.json"
            with open(tmp_path, "w") as f:
                json.dump(value, f)
        else:
            raise ValueError("Unsupported value type for upload.")

        # upload to bucket
        gcs_path = f"results/{self.project}/{self.dataset}/{self.experiment_run_name}/{tmp_path}"
        blob = self.pixaris_bucket.blob(gcs_path)
        blob.upload_from_filename(tmp_path)
        # put together the clickable link
        clickable_link = f"https://console.cloud.google.com/storage/browser/_details/{self.gcp_pixaris_bucket_name}/{gcs_path}?project={self.gcp_project_id}"
        # cleanup temporary file
        os.remove(tmp_path)
        return clickable_link

    def _add_default_metrics(self, bigquery_input: dict):
        """
        Adds default metrics to the BigQuery input dictionary if they are not already present.

        :param bigquery_input: The dictionary to update with default metrics.
        :type bigquery_input: dict
        """
        default_metrics = {
            "llm_reality": 0.0,
            "llm_similarity": 0.0,
            "llm_errors": 0.0,
            "llm_todeloy": 0.0,
            "iou": 0.0,
            "hyperparameters": "",
            "generation_params": "",
            "workflow_apiformat_json": "",
            "workflow_pillow_image": "",
            "max_parallel_jobs": 0.0,
        }
        for k, v in default_metrics.items():
            bigquery_input.setdefault(k, v)

    def _prepare_additional_pillow_images_upload(
        self, args: dict[str, any], image_name_pairs: Iterable[tuple[Image.Image, str]]
    ):
        """
        Prepares additional pillow images for upload by saving them under their node names in image_name_pairs.
        This function also removes the pillow_images key from args to prevent re-uploading.

        :param args: A dictionary containing additional arguments, including pillow images.
        :type args: dict[str, any]
        :param image_name_pairs: An iterable of tuples containing images and their names.
        :type image_name_pairs: Iterable[tuple[Image.Image, str]]
        """
        # all images in pillow_images are additional inputs by the user, e.g. inspo images. No dataset is in here.
        if pillow_images := args.get("pillow_images"):
            assert isinstance(pillow_images, list), (
                "pillow_images must be a list of dictionaries."
            )

            # each image in pillow_images needs to be saved under an own key in dict for easier saving
            for image_details in pillow_images:
                image_name = (
                    "z_" + image_details["node_name"].replace(" ", "_") + ".png"
                )
                pillow_image = image_details["pillow_image"]
                # save the image under its node_name
                image_name_pairs.append((pillow_image, image_name))

            # remove pillow_images from args so they are not uploaded later on again
            args.pop("pillow_images", None)

    def _store_experiment_parameters_and_results(
        self,
        metric_values: dict[str, float],
        args: dict[str, any] = {},
    ):
        """
        Stores experiment parameters and results in BigQuery and Google Cloud Storage.

        :param metric_values: A dictionary of metric names and their values.
        :type metric_values: dict[str, float]
        :param args: Additional arguments, such as images or JSON data.
        :type args: dict[str, any]
        """
        timestamp = time.strftime("%Y%m%d-%H%M%S")

        bigquery_input = {
            "timestamp": timestamp,
            "experiment_run_name": self.experiment_run_name,
        }

        for key, value in args.items():
            if isinstance(value, (Image.Image, dict)):  # E.g. for workflow images
                gcp_path = self._upload_to_gcs(key, value)
                bigquery_input[key] = gcp_path
            elif isinstance(value, int):
                bigquery_input[key] = value
            elif isinstance(value, float):
                bigquery_input[key] = value
            else:
                bigquery_input[key] = str(value)

        for key, value in metric_values.items():
            if isinstance(value, (dict)):  # E.g. for Hyperparameters
                gcp_path = self._upload_to_gcs(key, value)
                bigquery_input[key] = gcp_path
            elif isinstance(value, int):
                bigquery_input[key] = value
            elif isinstance(value, float):
                bigquery_input[key] = value
            else:
                bigquery_input[key] = str(value)

        # Ensure default metrics are present
        self._add_default_metrics(bigquery_input)

        # Define table reference
        table_ref = f"{self.gcp_bq_experiment_dataset}.{self.project}_{self.dataset}_experiment_results"

        # Ensure table exists with correct schema
        ensure_table_exists(
            table_ref=table_ref,
            bigquery_input=bigquery_input,
            bigquery_client=self.bigquery_client,
        )

        # Insert the row into BigQuery
        errors = self.bigquery_client.insert_rows_json(table_ref, [bigquery_input])

        # Check for errors and display warnings to UI
        if errors == []:
            print(f"Inserted row into table {table_ref}.")
        else:
            raise RuntimeError(f"Failed to insert row into table {table_ref}: {errors}")

    def _store_generated_images(
        self,
        image_name_pairs: Iterable[tuple[Image.Image, str]],
    ):
        """
        Store generated images in the Google Cloud Storage bucket.

        :param image_name_pairs: An iterable of tuples containing PIL Image objects and their corresponding names.
        :type image_name_pairs: Iterable[tuple[Image.Image, str]]
        """

        # Upload each image to the GCS bucket
        for pillow_image, name in image_name_pairs:
            image_path = f"{name}"
            gcp_image_path = f"results/{self.project}/{self.dataset}/{self.experiment_run_name}/generated_images/{name}"
            pillow_image.save(image_path)
            blob = self.pixaris_bucket.blob(gcp_image_path)
            blob.upload_from_filename(image_path)
            print(f"Uploaded {name} to {gcp_image_path}")
            os.remove(image_path)

    def store_results(
        self,
        project: str,
        dataset: str,
        experiment_run_name: str,
        image_name_pairs: Iterable[tuple[Image.Image, str]],
        metric_values: dict[str, float],
        args: dict[str, any] = {},
    ):
        """
        Stores the results of an experiment, including images, metrics, and parameters.

        :param project: The name of the project.
        :type project: str
        :param dataset: The name of the dataset.
        :type dataset: str
        :param experiment_run_name: The name of the experiment run.
        :type experiment_run_name: str
        :param image_name_pairs: An iterable of tuples containing images and their names.
        :type image_name_pairs: Iterable[tuple[Image.Image, str]]
        :param metric_values: A dictionary of metric names and their values.
        :type metric_values: dict[str, float]
        :param args: Additional arguments, such as images or JSON data.
        :type args: dict[str, any]
        """

        self.project = project
        self.dataset = dataset
        self.storage_client = storage.Client(project=self.gcp_project_id)
        self.bigquery_client = bigquery.Client(project=self.gcp_project_id)
        self.pixaris_bucket = self.storage_client.bucket(self.gcp_pixaris_bucket_name)

        # set and adjust experiment_run_name with timestamp
        self.experiment_run_name = experiment_run_name
        self.experiment_run_name = self._ensure_unique_experiment_run_name()
        # prevent that args["experiment_run_name"] will overwrite unique experiment_run_name
        args["experiment_run_name"] = self.experiment_run_name

        self._validate_args(args=args)

        # store images that were generated by the generator AND additional input images from args
        self._prepare_additional_pillow_images_upload(
            args=args, image_name_pairs=image_name_pairs
        )

        # upload all content of metric_values and args to bigquery and bucket if applicable
        self._store_experiment_parameters_and_results(
            metric_values=metric_values, args=args
        )

        self._store_generated_images(image_name_pairs=image_name_pairs)

    def load_projects_and_datasets(self) -> dict:
        """
        Loads the projects and datasets available in the Google Cloud Storage bucket.

        :return: A dictionary mapping project names to lists of dataset names.
        :rtype: dict
        """
        self.storage_client = storage.Client(project=self.gcp_project_id)
        self.pixaris_bucket = self.storage_client.bucket(self.gcp_pixaris_bucket_name)

        blobs = self.pixaris_bucket.list_blobs()

        project_dict = {}
        for blob in blobs:
            name = blob.name
            if name.startswith("results/") and name != "results/":
                prefix_removed = name.split("results/")[1]
                parts = prefix_removed.split("/")
                if (
                    len(parts) >= 2
                ):  # Ensure there is at least a project and dataset level
                    project, dataset = parts[0], parts[1]
                    if project not in project_dict and project != "pickled_results":
                        project_dict[project] = []
                    if (
                        dataset not in project_dict[project]
                        and dataset != "feedback_iterations"
                    ):
                        project_dict[project].append(dataset)
        return project_dict

    def load_experiment_results_for_dataset(
        self,
        project: str,
        dataset: str,
    ) -> pd.DataFrame:
        """
        Loads the results of an experiment from a BigQuery dataset.

        :param project: The name of the project.
        :type project: str
        :param dataset: The name of the dataset.
        :type dataset: str
        :return: The results of the experiment as a pandas DataFrame.
        :rtype: pd.DataFrame
        """
        query = f"""
        SELECT *
        FROM `{self.gcp_bq_experiment_dataset}.{project}_{dataset}_experiment_results`
        """

        self.bigquery_client = bigquery.Client(project=self.gcp_project_id)

        try:
            query_job = self.bigquery_client.query(query)
            results = query_job.result()
            return results.to_dataframe()
        except Exception as e:
            raise RuntimeError(f"Failed to load experiment results from BigQuery: {e}")

    def load_images_for_experiment(
        self,
        project: str,
        dataset: str,
        experiment_run_name: str,
        local_results_directory: str,
    ) -> list[str]:
        """
        Downloads images for a experiment_run_name from GCP bucket to local directory.
        Returns list of local image paths that belong to the experiment_run_name.

        :param project: The name of the project.
        :type project: str
        :param dataset: The name of the dataset.
        :type dataset: str
        :param experiment_run_name: Name of the experiment run.
        :type experiment_run_name: str
        :param local_results_directory: The local directory to store the downloaded images.
        :type local_results_directory: str
        :return: List of local image paths.
        :rtype: list[str]
        """
        self.storage_client = storage.Client(project=self.gcp_project_id)
        self.pixaris_bucket = self.storage_client.bucket(self.gcp_pixaris_bucket_name)

        print(f"Downloading images for experiment_run_name {experiment_run_name}...")
        path_in_parent_folder = (
            f"{project}/{dataset}/{experiment_run_name}/generated_images/"
        )
        # list images in bucket/project/dataset/experiment_run_name
        blobs = self.pixaris_bucket.list_blobs(
            prefix=f"results/{path_in_parent_folder}",
        )

        local_image_paths = []
        # download images
        for blob in blobs:
            if blob.name.endswith("/"):
                continue  # directory, skip.
            image_path_local = os.path.join(
                local_results_directory, blob.name.replace("results/", "")
            )
            local_image_paths.append(image_path_local)

            # download image if not already downloaded
            if not os.path.exists(image_path_local):
                gr.Info(
                    f"Downloading image '{blob.name.split('/')[-1]}'...", duration=1
                )
                os.makedirs(os.path.dirname(image_path_local), exist_ok=True)
                blob.download_to_filename(image_path_local)

        print("Done.")
        local_image_paths.sort()
        return local_image_paths
